import collections, functools, types, itertools, warnings, io
deque = collections.deque

__all__ = [ 'Reiterable', 'reiterable', 'reiterablemap', 'reiterablefilter', 'reiterableaccumulate', 'emptyReiterable', 'ReitLinesFiles', 'ReitDBTable' ]

emptyReiterable = lambda: reiterable(())

class Reiterable(collections.abc.Iterator):
	"""
	A reiterable iterator. To implement this class, one should
	provide at least the `__reiter__` method, that returns a reset
	of the iterator.

	Instance attributes:
	+	_source_start:
		a integer that indicates how many first elements of the
		iterable obtained by calling the `__reiter__` method should
		not be part of the instance underlying iterable.
	+	_source_count:
		a counter that indicates the current element index in the
		source iterable.
	+ _count:
		a counter that indicates the current element index in the
		underlying iterable.
	+	_args, _kwargs:
		respectively, the arguments list and keyworded parameters,
		that can be used in the `__reiter__` method.
	"""
	def __init__(self, *args, source_start=0, **kwargs):
		"""
		+	source_start:
			an index in the source iterable (namely, the one returned
			by `self.__reiter__(*self._args, **self._kwargs)`) from
			which to start (default is `0`).
		+	args, kwargs:
			respectively, the tuple of arguments and the dictionary of
			keyworded parameters to be pass to the `self.__reiter__`
			method.
		"""
		self._args = args
		self._kwargs = kwargs
		self._source_start = source_start
		self._source_count = source_start
		self._count = 0
		self._len = None
		self._iterator = self.reiter()
	def length(self):
		try:
			length = len(self)
		except TypeError:
			warnings.warn('Computing length of Reiterable {self} may enter an infinite loop')
			for length, _ in enumerate(self):
				pass
			else:
				length = 0
		return length
	def contains(self, k):
		try:
			return k in self
		except TypeError:
			warnings.warn('Checking membership of {k} in Reiterable {self} may enter an infinite loop')
			return any(v == k for v in self)
	@classmethod
	def build_from(cls, self):
		args = self._args
		kwargs = self._kwargs
		source_start = self._source_count
		res = self.__class__(*args, source_start=source_start, **kwargs)
		return res
	def __iter__(self):
		return self.build_from(self)
	def __eq__(self, other):
		if not 'reiter' in dir(other):
			return False
		if self._count != other._count:
			return False
		tmp = object()
		L = self.reiter()
		R = other.reiter()
		try:
			while True:
				l = next(L)
				r = next(R)
				assert l == r
				r = tmp
		except StopIteration:
			if r is tmp:
				return next(R, tmp) is tmp
			else:
				return False
		except AssertionError:
			return False
	def reset(self):
		self._iterator = self.reiter()
		self._source_count = self._source_start
		self._count = 0
	def __next__(self):
		return self.next()
	def next(self):
		assert self._source_count >= self._source_start
		val = next(self._iterator)
		self._source_count += 1
		orig_c = self._source_count
		self._count += 1
		return val
	def previous(self):
		if not self._count:
			raise StopIteration(f"Cannot apply previous on non-started iterable")
		orig_c = self._source_count
		self.reset()
		for _ in range(orig_c-1):
			val = next(self)
		return val
	def __bool__(self):
		return not self.is_empty()
	def __getitem__(self, k):
		if isinstance(k, slice):
			start, stop, step = k.start, k.stop, k.step
			return self.getslice(start, stop, step)
		else:
			res = self[k:]
			return next(res)
	@functools.wraps(slice)
	def getslice(self, *args):
		if len(args) == 1 and isinstance(args[0], slice):
			s = args[0]
		else:
			s = slice(*args)
		start = 0 if s.start is None else s.start
		step = 1 if s.step is None else s.step
		stop = s.stop
		if step==0:
			raise ValueError(f"slice step cannot be 0")
		if step<0:
			raise NotImplementedError
		if stop==0 or (stop is not None and (0<stop<=start or stop<=start<0)):
			return emptyReiterable()
		if start < 0:
			#length might enter an infinite loop if `self` is infinite
			length = self.length()
			start = max(0, length + start)
			if stop is not None and stop < 0:
				stop = max(0, length + stop)
				if stop == 0:
					return emptyReiterable()
		#start>=0, step>=0, stop?
		res = iter(self)
		@reiterable
		def sliced_reiterable(start, stop, step):
			res.reset()
			pos = 0
			try:
				while True:
					if pos>=start:
						if stop is not None and stop<0:
							res[-stop]
						yield next(res)
						for _ in range(step-1):
							next(res)
							pos += 1
					else:
						next(res)
					pos += 1
					if stop is not None and stop>0 and pos>=stop:
						return
			except (StopIteration, IndexError):
				return
		return sliced_reiterable(start, stop, step)
	def is_empty(self):
		res = iter(self)
		try:
			next(res)
			return False
		except StopIteration:
			return True
	def map(self, func):
		res = self.build_from(self)
		#TODO: decorator of map
		@reiterable
		def g():
			res.reset()
			yield from map(func, res)
		return g()
	def filter(self, func):
		res = self.build_from(self)
		@reiterable
		def g():
			res.reset()
			yield from filter(func, res)
		return g()
	def accumulate(self, func, initial=0):
		res = self.build_from(self)
		@reiterable
		def g():
			res.reset()
			acc = initial
			for i in res:
				acc = func(acc, i)
				yield acc
		return g()
	def reduce(self, func, initial=0):
		try:
			len(self)
		except TypeError:
			warnings.warn('Reducing Reiterable {self} may enter an infinite loop')
		return functools.reduce(func, iter(self), initial=initial)
	def reiter(self, skip=0):
		"""
		A wrapper for `__reiter__` which returns a reset copy of the
		underlying iterator, from which the first items are popped
		off.
		"""
		res = iter(self.__reiter__(*self._args, **self._kwargs))
		for _ in range(self._source_start+skip):
			next(res)
		return res
	def __reiter__(self, *args, **kwargs):
		raise NotImplementedError
	def cycle(self, marker=None):
		"""
		Returns a new resettable iterator, that cycles whence never
		raises `StopIteration` exception. If `marker` is not `None`
		then it is yielded each time the iterator restarts, namely
		between the last element of `self`'s underlying iterator and
		its first one. This behavior can be changed by resetting the
		`_marker` attribute of the resulting reiterable iterator to
		`None` after.
		"""
		raise NotImplementedError

class SequenceReiterable(Reiterable, collections.abc.Sequence):
	def __len__(self):
		return self.length()
	def __contains__(self, k):
		return self.contains()

#decorators
def reiterableGenerator(gen):
	"""
	A decorator for reiterable generators.
	"""
	def __reiter__(self, *args, **kwargs):
		return gen(*args, **kwargs)
	def send(self, *args):
		return self._iterator.send(*args)
	def throw(self, *args):
		return self._iterator.throws(*args)
	meth = dict(__reiter__=__reiter__, send=send, throw=throw)
	ReiterableGenerator = type('ReiterableGenerator', (Reiterable, collections.abc.Generator), meth)
	return ReiterableGenerator

def reiterableIterable(iterable):
	"""
	A decorator for iterables. Here, `iterable` can be either an
	iterable class (e.g., implementing `collections.abc.Iterable`)
	or an iterable instance.
	"""
	is_instance = not isinstance(iterable, type)
	if is_instance:
		assert iter(iterable) is not iterable
		cls = iterable.__class__
	else:
		cls = iterable
	meth = {}
	if '__reiter__' not in dir(iterable):
		if is_instance:
			def __reiter__(self):
				return iter(iterable)
		else:
			__reiter__ = cls
		meth['__reiter__'] = __reiter__
	#Override sequence methods
	if '__len__' in dir(iterable):
		bases = (SequenceReiterable,)
		def __len__(self):
			return (len(iterable) - self._source_count)
		meth['__len__'] = __len__
	else:
		bases = (Reiterable,)
	if '__getitem__' in dir(iterable):
		def __getitem__(self, k):
			try:
				return iterable[k+self._source_count]
			except TypeError:
				return bases[0].__getitem__(self, k)
		meth['__getitem__'] = __getitem__
	#Base classes
	ReiterableIterator = type('ReiterableIterator', bases, meth)
	if is_instance:
		return ReiterableIterator()
	else:
		return ReiterableIterator

#main decorator
def reiterable(arg):
	if isinstance(arg, Reiterable):
		return arg.build_from(arg)
	elif isinstance(arg, types.FunctionType):
		return reiterableGenerator(arg)
	elif isinstance(arg, type):
		return reiterableIterable(arg)
	else:
		return reiterableIterable(arg)

#decored map, filter, accumulate
reiterablemap = Reiterable.map
reiterablefilter = Reiterable.filter
reiterableaccumulate = Reiterable.accumulate

@reiterable
def ReitLinesFiles(path):
	with open(path,"r") as f:
		for l in f:
			yield l.strip()
@reiterable
def ReitDBTable(db, query):
	c = db.cursor()
	c.execute(query)
	while True:
		try:
			t = c.fetchone()
			if t:
				yield t
			else:
				c.close()
				return
		except:
			c.close()
			return
